
class Node {
    int value;
    Node next;
}

/*@

predicate node(Node no, int v, Node ne)
	= no.value |-> v &*& no.next |-> ne;

@*/

/*@

inductive intlist = | my_nil | my_cons(int, intlist);

predicate llist(LList l, intlist a)
	= l.head |-> ?np &*& list(np, a);
	
predicate list(Node n, intlist a)
	= n == null ? a == my_nil
	         : node(n, ?i, ?ne) &*& list(ne, ?as) &*& a == my_cons(i, as);
	         
@*/

/*@

fixpoint intlist app(intlist l1, intlist l2) {
  switch (l1) {
    case my_nil: return l2;
    case my_cons(x, v): return my_cons(x, app(v, l2));
  }
}

fixpoint intlist rev(intlist l) {
  switch (l) {
    case my_nil: return my_nil;
    case my_cons(x, v): return app(rev(v), my_cons(x, my_nil));
  }
}

lemma void app_assoc(intlist l1, intlist l2, intlist l3)
  requires emp;
  ensures app(app(l1, l2), l3) == app(l1, app(l2, l3));
{
  switch (l1) {
    case my_nil:
    case my_cons(x, v):
      app_assoc(v, l2, l3);
  }
}

lemma void rev_lemma(int i, intlist l1, intlist l2)
  requires emp;
  ensures app(rev(my_cons(i, l1)), l2) == app(rev(l1), my_cons(i, l2));
{
  app_assoc(rev(l1), my_cons(i, my_nil), l2);
}

lemma void app_to_my_nil(intlist l)
  requires emp;
  ensures app(l, my_nil) == l;
{
  switch (l) {
    case my_nil:
    case my_cons(x, v):
      app_to_my_nil(v);
  }
}

lemma void rev_app(intlist l1, intlist l2)
  requires emp;
  ensures rev(app(l1, l2)) == app(rev(l2), rev(l1));
{
  switch (l1) {
    case my_nil:
      app_to_my_nil(rev(l2));
    case my_cons(x, v):
      rev_app(v, l2);
      app_assoc(rev(l2), rev(v), my_cons(x, my_nil));
  }
}

lemma void rev_twice(intlist l)
  requires emp;
  ensures rev(rev(l)) == l;
{
  switch (l) {
    case my_nil:
    case my_cons(x, v):
      rev_app(rev(v), my_cons(x, my_nil));
      rev_twice(v);
  }
}

@*/

class LList {

    void reverse()
        //@ requires llist(this, ?a);
        //@ ensures llist(this, rev(a));
    {
        Node previous = null;
        //@ open llist(this, a);
        Node next = this.head;
        Node temp = null;
        this.head = null;
        //@ close list(null, my_nil);
        while (next != null)
            //@ invariant this.head |-> ?n &*& list(n, ?b) &*& list(next, ?g) &*& a == app(rev(b), g);
        {
            previous = this.head;
            this.head = next;
            temp = next;
            //@ open list(temp, g);
            //@ open node(temp, ?i, _);
            next = temp.next;
            temp.next = previous;
            // The following two lines get a handle on the tail of g:
            //@ open list(next, ?gs);
            //@ close list(next, gs);
            //@ close node(temp, i, previous);
            //@ close list(temp, my_cons(i, b));
            //@ rev_lemma(i, b, gs);
        }
        //@ open list(next, _);
        //@ app_to_my_nil(rev(b));
        //@ rev_twice(b);
        //@ close llist(this, rev(a));
    }   
    
    private Node head;
}
